/*
 * Copyright (c) 1995 Picture Elements
 *	Stephen Williams (steve@icarus.com)
 *	Gus Baldauf (gbaldauf@ix.netcom.com)
 */
#ident "$Id: bpp.c,v 1.3 1996/06/02 16:43:40 steve Exp $"
/*
 * $Log: bpp.c,v $
 * Revision 1.3  1996/06/02 16:43:40  steve
 *  Move to 1.3 kernel.
 *
 * Revision 1.2  1995/11/09 00:33:15  steve
 *  More thorough cleanup if terminate fails.
 *
 * Revision 1.1  1995/10/27 03:22:57  steve
 *  Add bpp to CVS.
 *
 *
 */


# include  <linux/kernel.h>
# include  <linux/module.h>
# include  <linux/version.h>
# include  <linux/fs.h>
# include  <linux/errno.h>
# include  <linux/sched.h>
# include  <linux/timer.h>
# include  <linux/ioport.h>
# include  <asm/io.h>
# include  <asm/system.h>
# include  <asm/segment.h>
# include  "bpp.h"

#define BPP_PROBE_CODE 0x55
#define BPP_DELAY 100

const unsigned  BPP_MAJOR = 32;
const char* dev_name = "bpp";

/* When switching from compatability to a mode where I can read, try
   the following mode first. */

/* const unsigned char DEFAULT_ECP = 0x10; */
const unsigned char DEFAULT_ECP = 0x30;
const unsigned char DEFAULT_NIBBLE = 0x00;

/*
 * These are 1284 time constraints, in units of jiffies.
 */

const unsigned long TIME_PSetup = 1;
const unsigned long TIME_PResponse = 6;
const unsigned long TIME_IDLE_LIMIT = 2000;

/*
 * One instance per supported subdevice...
 */
# define BPP_NO 3

enum IEEE_Mode { COMPATIBILITY, NIBBLE, ECP, ECP_RLE, EPP };

struct inst {
      unsigned present  : 1; /* True if the hardware exists */
      unsigned enhanced : 1; /* True if the hardware in "enhanced" */
      unsigned opened   : 1; /* True if the device is opened already */
      unsigned run_flag : 1; /* True if waiting for a repeate byte */

      unsigned char direction; /* 0 --> out, 0x20 --> IN */
      unsigned char pp_state; /* State of host controlled pins. */
      enum IEEE_Mode mode;

      unsigned char run_length;
      unsigned char repeat_byte;

      /* These members manage timeouts for programmed delays */
      struct wait_queue *wait_queue;
      struct timer_list timer_list;
};

static struct inst instances[BPP_NO];
const unsigned short base_addrs[BPP_NO] = { 0x278, 0x378, 0x3bc };

/*
 * This is TRUE if the module_init successfully loaded the module.
 */

static int loaded_flag = 0;

/*
 * This method takes the pin values mask and sets the hardware pins to
 * the requested value: 1 == high voltage, 0 == low voltage. This
 * burries the annoying PC bit inversion and preserves the direction
 * flag.
 */
static void set_pins(unsigned short pins, unsigned minor)
{
      unsigned char bits = instances[minor].direction;

      if (! (pins & BPP_PP_nStrobe))   bits |= 1;
      if (! (pins & BPP_PP_nAutoFd))   bits |= 2;
      if (   pins & BPP_PP_nInit)      bits |= 4;
      if (! (pins & BPP_PP_nSelectIn)) bits |= 8;

      instances[minor].pp_state = bits;

      outb_p(bits, base_addrs[minor]+2);
}

static unsigned short get_pins(unsigned minor)
{
      unsigned short bits = 0;

      unsigned value = instances[minor].pp_state;
      if (! (value & 0x01)) bits |= BPP_PP_nStrobe;
      if (! (value & 0x02)) bits |= BPP_PP_nAutoFd;
      if (value & 0x04)     bits |= BPP_PP_nInit;
      if (! (value & 0x08)) bits |= BPP_PP_nSelectIn;

      value = inb_p(base_addrs[minor]+1);
      if (value & 0x08)     bits |= BPP_GP_nFault;
      if (value & 0x10)     bits |= BPP_GP_Select;
      if (value & 0x20)     bits |= BPP_GP_PError;
      if (value & 0x40)     bits |= BPP_GP_nAck;
      if (! (value & 0x80)) bits |= BPP_GP_Busy;

      return bits;
}

static void bpp_wake_up(unsigned long val)
{ wake_up(&instances[val].wait_queue); }

static void snooze(unsigned long snooze_time, unsigned minor)
{
      instances[minor].timer_list.expires = snooze_time+1;
      instances[minor].timer_list.data    = minor;
      add_timer(&instances[minor].timer_list);
      sleep_on (&instances[minor].wait_queue);
}

static int wait_for(unsigned short set, unsigned short clr,
		   unsigned long delay, unsigned minor)
{
      unsigned short pins = get_pins(minor);

      unsigned long extime = 0;

      /*
       * Try a real fast scan for the first jiffy, in case the device
       * responds real good. The first while loop guesses an expire
       * time accounting for possible wraparound of jiffies.
       */
      while (extime <= jiffies) extime = jiffies + 1;
      while ( (jiffies < extime)
	     && (((pins & set) != set) || ((pins & clr) != 0)) ) {
	    pins = get_pins(minor);
      }

      delay -= 1;

      /*
       * If my delay expired or the pins are still not where I want
       * them, then resort to using the timer and greatly reduce my
       * sample rate. If the peripheral is going to be slow, this will
       * give the CPU up to some more worthy process.
       */
      while ( delay && (((pins & set) != set) || ((pins & clr) != 0)) ) {

	    snooze(1, minor);
	    pins = get_pins(minor);
	    delay -= 1;
      }

      if (delay == 0) return -1;
      else return pins;
}

/*
 * Return ZERO(0) If the negotiation succeeds, an errno otherwise. An
 * errno means something broke, and I do not yet know how to fix it.
 */
static int negotiate(unsigned char mode, unsigned minor)
{
      int rc;
      unsigned short pins = get_pins(minor);
      if (pins & BPP_PP_nSelectIn) return -EIO;


        /* Event 0: Write the mode to the data lines */
      outb_p(mode, base_addrs[minor]);

      snooze(TIME_PSetup, minor);

        /* Event 1: Strobe the mode code into the peripheral */
      set_pins(BPP_PP_nSelectIn|BPP_PP_nStrobe|BPP_PP_nInit, minor);

        /* Wait for Event 2: Peripheral responds as a 1284 device. */
      rc = wait_for(BPP_GP_PError|BPP_GP_Select|BPP_GP_nFault,
		    BPP_GP_nAck,
		    TIME_PResponse,
		    minor);

      if (rc == -1) return -ETIMEDOUT;

        /* Event 3: latch extensibility request */
      set_pins(BPP_PP_nSelectIn|BPP_PP_nInit, minor);

        /* ... quick nap while peripheral ponders the byte i'm sending...*/
      snooze(1, minor);

        /* Event 4: restore strobe, to ACK peripheral's response. */
      set_pins(BPP_PP_nSelectIn|BPP_PP_nAutoFd|BPP_PP_nStrobe|BPP_PP_nInit, minor);

        /* Wait for Event 6: Peripheral latches response bits */
      rc = wait_for(BPP_GP_nAck, 0, TIME_PSetup+TIME_PResponse, minor);
      if (rc == -1) return -EIO;

        /* A 1284 device cannot refuse nibble mode */
      if (mode == DEFAULT_NIBBLE) return 0;

      if (pins & BPP_GP_Select) return 0;

      return -EPROTONOSUPPORT;
}

static int terminate(unsigned minor)
{
      int rc;

        /* Event 22: Request termination of 1284 mode */
      set_pins(BPP_PP_nAutoFd|BPP_PP_nStrobe|BPP_PP_nInit, minor);

        /* Wait for Events 23 and 24: ACK termination request. */
      rc = wait_for(BPP_GP_Busy|BPP_GP_nFault,
		    BPP_GP_nAck,
		    TIME_PSetup+TIME_PResponse,
		    minor);

      instances[minor].direction = 0;
      instances[minor].mode = COMPATIBILITY;

      if (rc == -1) {
	    return -EIO;
      }

        /* Event 25: Handshake by lowering nAutoFd */
      set_pins(BPP_PP_nStrobe|BPP_PP_nInit, minor);

        /* Event 26: Peripheral wiggles lines... */

        /* Event 27: Peripheral sets nAck HIGH to ack handshake */
      rc = wait_for(BPP_GP_nAck, 0, TIME_PResponse, minor);
      if (rc == -1) {
	    set_pins(BPP_PP_nAutoFd|BPP_PP_nStrobe|BPP_PP_nInit, minor);
	    return -EIO;
      }

        /* Event 28: Finish phase by raising nAutoFd */
      set_pins(BPP_PP_nAutoFd|BPP_PP_nStrobe|BPP_PP_nInit, minor);

      return 0;
}


/*
 * Allow only one process to open the device at a time.
 */
static int open(struct inode *inode, struct file *f)
{
      unsigned minor = MINOR(inode->i_rdev);
      if (minor >= BPP_NO) return -ENODEV;
      if (! instances[minor].present) return -ENODEV;
      if (instances[minor].opened) return -EBUSY;

      instances[minor].opened = 1;

      return 0;
}

/*
 * When the process closes the device, this method is called to clean
 * up and reset the hardware. Always leave the device in compatibility
 * mode as this is a reasonable place to clean up from messes made by
 * ioctls, or other mayhem.
 */
static void release(struct inode *inode, struct file *f)
{
      unsigned minor = MINOR(inode->i_rdev);
      instances[minor].opened = 0;

      if (instances[minor].mode != COMPATIBILITY)
	terminate(minor);
}

static int read_nibble(unsigned minor, char*c, int cnt)
{
      unsigned remaining = cnt;
      int rc;

      while (remaining > 0) {
	    unsigned char byte = 0;
	    int pins;

	    /* Event 7: request nibble */
	    set_pins(BPP_PP_nSelectIn|BPP_PP_nStrobe, minor);

	    /* Wait for event 9: Peripher strobes first nibble */
	    pins = wait_for(0, BPP_GP_nAck, TIME_IDLE_LIMIT, minor);
	    if (pins == -1) return -ETIMEDOUT;

	    /* Event 10: I handshake nibble */
	    set_pins(BPP_PP_nSelectIn|BPP_PP_nStrobe|BPP_PP_nAutoFd, minor);
	    if (pins & BPP_GP_nFault) byte |= 0x01;
	    if (pins & BPP_GP_Select) byte |= 0x02;
	    if (pins & BPP_GP_PError) byte |= 0x04;
	    if (pins & BPP_GP_Busy)   byte |= 0x08;

	    /* Wait for event 11: Peripheral handshakes nibble */
	    rc = wait_for(BPP_GP_nAck, 0, TIME_PResponse, minor);

	    /* Event 7: request nibble */
	    set_pins(BPP_PP_nSelectIn|BPP_PP_nStrobe, minor);

	    /* Wait for event 9: Peripher strobes first nibble */
	    pins = wait_for(0, BPP_GP_nAck, TIME_PResponse, minor);
	    if (rc == -1) return -ETIMEDOUT;

	    /* Event 10: I handshake nibble */
	    set_pins(BPP_PP_nSelectIn|BPP_PP_nStrobe|BPP_PP_nAutoFd, minor);
	    if (pins & BPP_GP_nFault) byte |= 0x10;
	    if (pins & BPP_GP_Select) byte |= 0x20;
	    if (pins & BPP_GP_PError) byte |= 0x40;
	    if (pins & BPP_GP_Busy)   byte |= 0x80;

	    put_user(byte, c);
	    c += 1;
	    remaining -= 1;

	    /* Wait for event 11: Peripheral handshakes nibble */
	    rc = wait_for(BPP_GP_nAck, 0, TIME_PResponse, minor);
	    if (rc == -1) return -EIO;
      }

      return cnt - remaining;
}

static int read_ecp(unsigned minor, char *c, int cnt)
{
      unsigned remaining;
      int rc;

        /* Turn ECP mode from forward to reverse if needed. */
      if (! instances[minor].direction) {
	    unsigned short pins = get_pins(minor);

	      /* Event 38: Turn the bus around */
	    instances[minor].direction = 0x20;
	    pins &= ~BPP_PP_nAutoFd;
	    set_pins(pins, minor);

	      /* Event 39: Set pins for reverse mode. */
	    snooze(TIME_PSetup, minor);
	    set_pins(BPP_PP_nStrobe|BPP_PP_nSelectIn, minor);

	      /* Wait for event 40: Peripheral ready to be strobed */
	    rc = wait_for(0, BPP_GP_PError, TIME_PResponse, minor);
	    if (rc == -1) return -ETIMEDOUT;
      }

      remaining = cnt;

      while (remaining > 0) {

	      /* If there is a run length for a repeated byte, repeat */
	      /* that byte a few times. */
	    if (instances[minor].run_length && !instances[minor].run_flag) {

		  char buffer[128];
		  unsigned idx;
		  unsigned repeat = remaining < instances[minor].run_length
		                         ? remaining
					 : instances[minor].run_length;

		  for (idx = 0 ;  idx < repeat ;  idx += 1)
		    buffer[idx] = instances[minor].repeat_byte;

		  memcpy_tofs(c, buffer, repeat);
		  remaining -= repeat;
		  c += repeat;
		  instances[minor].run_length -= repeat;
	    }

	    if (remaining == 0) break;


	      /* Wait for Event 43: Data active on the bus. */
	    rc = wait_for(0, BPP_GP_nAck, TIME_IDLE_LIMIT, minor);
	    if (rc == -1) break;

	    if (rc & BPP_GP_Busy) {
		    /* OK, this is data. read it in. */
		  unsigned char byte = inb(base_addrs[minor]);
		  put_user(byte, c);
		  c += 1;
		  remaining -= 1;

		  if (instances[minor].run_flag) {
			instances[minor].repeat_byte = byte;
			instances[minor].run_flag = 0;
		  }

	    } else {
		  unsigned char byte = inb(base_addrs[minor]);
		  if (byte & 0x80) {
			printk("bpp%d: "
			       "Ignoring ECP channel %u from device.\n",
			       minor, byte & 0x7f);
		  } else {
			instances[minor].run_length = byte;
			instances[minor].run_flag = 1;
		  }
	    }

	      /* Event 44: I got it. */
	    set_pins(BPP_PP_nStrobe|BPP_PP_nAutoFd|BPP_PP_nSelectIn, minor);

	      /* Wait for event 45: peripheral handshake */
	    rc = wait_for(BPP_GP_nAck, 0, TIME_PResponse, minor);
	    if (rc == -1) return -ETIMEDOUT;

 	      /* Event 46: Finish handshake */
	    set_pins(BPP_PP_nStrobe|BPP_PP_nSelectIn, minor);

      }


      return cnt - remaining;
}

static int read(struct inode *inode, struct file *f, char *c, int cnt)
{
      int rc;
      const unsigned minor = MINOR(inode->i_rdev);
      if (minor >= BPP_NO) return -ENODEV;
      if (!instances[minor].present) return -ENODEV;

      switch (instances[minor].mode) {

	  default:
	    if (instances[minor].mode != COMPATIBILITY)
	      terminate(minor);

	    if (instances[minor].enhanced) {
		  /* For now, do all reads with ECP-RLE mode */
		  unsigned short pins;

		  rc = negotiate(DEFAULT_ECP, minor);
		  if (rc < 0) break;

		  instances[minor].mode = ECP_RLE;

		  /* Event 30: set nAutoFd low to setup for ECP mode */
		  pins = get_pins(minor);
		  pins &= ~BPP_PP_nAutoFd;
		  set_pins(pins, minor);

		  /* Wait for Event 31: peripheral ready */
		  rc = wait_for(BPP_GP_PError, 0, TIME_PResponse, minor);
		  if (rc == -1) return -ETIMEDOUT;

		  rc = read_ecp(minor, c, cnt);

	    } else {
		  rc = negotiate(DEFAULT_NIBBLE, minor);
		  if (rc < 0) break;

		  instances[minor].mode = NIBBLE;

		  rc = read_nibble(minor, c, cnt);
	    }
	    break;

	  case NIBBLE:
	    rc = read_nibble(minor, c, cnt);
	    break;

	  case ECP:
	  case ECP_RLE:
	    rc = read_ecp(minor, c, cnt);
	    break;

      }


      return rc;
}

/*
 * Compatibility mode handshaking is a matter of writing data,
 * strobing it, and waiting for the printer to stop being busy.
 */
static int write_compat(unsigned minor, char*c, int cnt)
{
      int rc;
      unsigned short pins = get_pins(minor);

      unsigned remaining = cnt;

      while (remaining > 0) {
	    unsigned char byte = get_user(c);
	    c += 1;

	    rc = wait_for(BPP_GP_nAck, BPP_GP_Busy, TIME_IDLE_LIMIT, minor);
	    if (rc == -1) return -ETIMEDOUT;

	    outb_p(byte, base_addrs[minor]);
	    remaining -= 1;
	    /* snooze(1, minor); */

	    pins &= ~BPP_PP_nStrobe;
	    set_pins(pins, minor);

	    rc = wait_for(BPP_GP_Busy, 0, TIME_PResponse, minor);

	    pins |= BPP_PP_nStrobe;
	    set_pins(pins, minor);
      }

      return cnt - remaining;
}

/*
 * Write data using ECP mode. Watch out that the port may be set up
 * for reading. If so, turn the port around.
 */
static int write_ecp(unsigned minor, char*c, int cnt)
{
      unsigned short pins = get_pins(minor);
      unsigned remaining = cnt;

      if (instances[minor].direction) {
	    int rc;

	      /* Event 47 Request bus be turned around */
	    pins |= BPP_PP_nInit;
	    set_pins(pins, minor);

	      /* Wait for Event 49: Peripheral relinquished bus */
	    rc = wait_for(BPP_GP_PError, 0, TIME_PResponse, minor);

	    pins |= BPP_PP_nAutoFd;
	    instances[minor].direction = 0;
	    set_pins(pins, minor);
      }

      while (remaining > 0) {
	    unsigned char byte = get_user(c);
	    int rc = wait_for(0, BPP_GP_Busy, TIME_PResponse, minor);
	    if (rc == -1) return -ETIMEDOUT;

	    c += 1;

	    outb_p(byte, base_addrs[minor]);

	    pins &= ~BPP_PP_nStrobe;
	    set_pins(pins, minor);

	    pins |= BPP_PP_nStrobe;
	    rc = wait_for(BPP_GP_Busy, 0, TIME_PResponse, minor);
	    if (rc == -1) return -EIO;

	    set_pins(pins, minor);
      }

      return cnt - remaining;
}

/*
 * Write to the peripheral. Be sensitive of the current mode. If I'm
 * in a mode that can be turned around (ECP) then just do
 * that. Otherwise, terminate and do my writing in compat mode. This
 * is the safest course as any device can handle it.
 */
static int write(struct inode *inode, struct file *f, char *c, int cnt)
{
      int errno = 0;
      unsigned minor = MINOR(inode->i_rdev);
      if (minor >= BPP_NO) return -ENODEV;
      if (!instances[minor].present) return -ENODEV;

      switch (instances[minor].mode) {

	  case ECP:
	  case ECP_RLE:
	    errno = write_ecp(minor, c, cnt);
	    break;
	  case COMPATIBILITY:
	    errno = write_compat(minor, c, cnt);
	    break;
	  default:
	    terminate(minor);
	    errno = write_compat(minor, c, cnt);
      }

      return errno;
}

static int ioctl(struct inode *inode, struct file *f, unsigned int cmd,
		 unsigned long arg)
{
      int errno = 0;

      unsigned minor = MINOR(inode->i_rdev);
      if (minor >= BPP_NO) return -ENODEV;
      if (!instances[minor].present) return -ENODEV;


      switch (cmd) {

	  case BPP_PUT_PINS:
	    set_pins(arg, minor);
	    break;

	  case BPP_GET_PINS:
	    errno = get_pins(minor);
	    break;

	  case BPP_PUT_DATA:
	    outb_p(arg, base_addrs[minor]);
	    break;

	  case BPP_GET_DATA:
	    errno = inb_p(base_addrs[minor]);
	    break;

	  case BPP_SET_INPUT:
	    if (arg)
	      if (instances[minor].enhanced) {
		    unsigned short bits = get_pins(minor);
		    instances[minor].direction = 0x20;
		    set_pins(bits, minor);
	      } else {
		    errno = -ENOTTY;
	      }
	    else {
		  unsigned short bits = get_pins(minor);
		  instances[minor].direction = 0x00;
		  set_pins(bits, minor);
	    }
	    break;

	  default:
	    errno = -EINVAL;
      }

      return errno;
}

static struct file_operations fileops = {
      0,
      read, /* read */
      write, /* write */
      0,
      0,
      ioctl,
      0, /* mmap */
      open,
      release,
      0,
      0,
      0,
      0
};

static void probeLptPort(unsigned idx)
{
	unsigned int testvalue;
	const unsigned short lpAddr = base_addrs[idx];

	instances[idx].present = 0;
	instances[idx].enhanced = 0;
	instances[idx].direction = 0;
	instances[idx].mode = COMPATIBILITY;
	instances[idx].wait_queue = 0;
	instances[idx].run_length = 0;
	instances[idx].run_flag = 0;
	init_timer(&instances[idx].timer_list);
	instances[idx].timer_list.function = bpp_wake_up;
	if (check_region(lpAddr,3)) return;

	/*
	 * First, make sure the instance exists. Do this by writing to
	 * the data latch and reading the value back. If the port *is*
	 * present, test to see if it supports extended-mode
	 * operation. This will be required for IEEE1284 reverse
	 * transfers.
	 */
	outb_p(BPP_PROBE_CODE, lpAddr);
	for (testvalue=0; testvalue<BPP_DELAY; testvalue++)
		;
	testvalue = inb_p(lpAddr);
	if (testvalue == BPP_PROBE_CODE) {
	        unsigned save;
		instances[idx].present = 1;

		request_region(lpAddr,3, dev_name);
		save = inb_p(lpAddr+2);
		for (testvalue=0; testvalue<BPP_DELAY; testvalue++)
			;
		outb_p(save|0x20, lpAddr+2);
		for (testvalue=0; testvalue<BPP_DELAY; testvalue++)
			;
		outb_p(~BPP_PROBE_CODE, lpAddr);
		for (testvalue=0; testvalue<BPP_DELAY; testvalue++)
			;
		testvalue = inb_p(lpAddr);
		if ((testvalue&0xff) == (0xff&~BPP_PROBE_CODE))
			instances[idx].enhanced = 0;
		else
			instances[idx].enhanced = 1;
		outb_p(save, lpAddr+2);
	}


	/*
	 * Leave the port in compat idle mode.
	 */
	set_pins(BPP_PP_nAutoFd|BPP_PP_nStrobe|BPP_PP_nInit, idx);
}

int init_module()
{
      int rc;
      unsigned idx;


      for (idx = 0 ;  idx < BPP_NO ;  idx += 1) {
	    instances[idx].opened = 0;
	    probeLptPort(idx);

	    if (instances[idx].present)
	      printk("bpp%d: Port at 0x%03x: Enhanced mode %s\n", idx,
		     base_addrs[idx],
		     instances[idx].enhanced? "SUPPORTED" : "UNAVAILABLE");
      }

      rc = register_chrdev(BPP_MAJOR, dev_name, &fileops);
      if (rc == 0) loaded_flag = 1;

      return 0;
}

void cleanup_module()
{
      int rc;

      if (loaded_flag) {
	    unsigned idx;

	    rc = unregister_chrdev(BPP_MAJOR, dev_name);

	    for (idx = 0 ;  idx < BPP_NO ;  idx += 1) {
		  if (instances[idx].present)
		    release_region(base_addrs[idx], 3);
	    }
      }
}
